@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -
 #
-# This file is part of couchdbkit released under the MIT license. 
+# This file is part of couchdbkit released under the MIT license.
 # See the NOTICE for more information.
 
 """
@@ -30,26 +30,29 @@ Example:
 UNKOWN_INFO = {}
 
 
-import base64
-import cgi
+from collections import deque
 from itertools import groupby
 from mimetypes import guess_type
-import re
 import time
-import urlparse
-import warnings
 
 from restkit.util import url_quote
 
-from couchdbkit.exceptions import *
-import couchdbkit.resource as resource
-from couchdbkit.utils import validate_dbname, json
+from .exceptions import InvalidAttachment, NoResultFound, \
+ResourceNotFound, ResourceConflict, BulkSaveError, MultipleResultsFound
+from . import resource
+from .utils import validate_dbname
 
 
 DEFAULT_UUID_BATCH_COUNT = 1000
 
 def _maybe_serialize(doc):
     if hasattr(doc, "to_json"):
+        # try to validate doc first
+        try:
+            doc.validate()
+        except AttributeError:
+            pass
+
         return doc.to_json(), True
     elif isinstance(doc, dict):
         return doc.copy(), False
@@ -61,10 +64,13 @@ class Server(object):
     A Server object can be used like any `dict` object.
     """
 
+    resource_class = resource.CouchdbResource
+
     def __init__(self, uri='http://127.0.0.1:5984',
-            uuid_batch_count=DEFAULT_UUID_BATCH_COUNT, resource_instance=None,
-            pool_instance=None,
-            filters=None):
+            uuid_batch_count=DEFAULT_UUID_BATCH_COUNT,
+            resource_class=None, resource_instance=None,
+            **client_opts):
+
         """ constructor for Server object
 
         @param uri: uri of CouchDb host
@@ -83,19 +89,19 @@ class Server(object):
         self.uuid_batch_count = uuid_batch_count
         self._uuid_batch_count = uuid_batch_count
 
-        if resource_instance and isinstance(resource_instance, 
+        if resource_class is not None:
+            self.resource_class = resource_class
+
+        if resource_instance and isinstance(resource_instance,
                                 resource.CouchdbResource):
             resource_instance.initial['uri'] = uri
             self.res = resource_instance.clone()
-            if pool_instance is not None:
-                self.res.client_opts['pool_instance'] = pool_instance
-                
+            if client_opts:
+                self.res.client_opts.update(client_opts)
         else:
-            self.res = resource.CouchdbResource(uri, 
-                                pool_instance=pool_instance,
-                                filters=filters)
-        self._uuids = []
-        
+            self.res = self.resource_class(uri, **client_opts)
+        self._uuids = deque()
+
     def info(self):
         """ info of server
 
@@ -104,9 +110,9 @@ class Server(object):
         """
         try:
             resp = self.res.get()
-        except Exception, e:
+        except Exception:
             return UNKOWN_INFO
-        
+
         return resp.json_body
 
     def all_dbs(self):
@@ -115,30 +121,34 @@ class Server(object):
         """
         return self.res.get('/_all_dbs').json_body
 
+    def get_db(self, dbname, **params):
+        """
+        Try to return a Database object for dbname.
+
+        """
+        return Database(self._db_uri(dbname), server=self, **params)
+
     def create_db(self, dbname, **params):
         """ Create a database on CouchDb host
 
         @param dname: str, name of db
         @param param: custom parameters to pass to create a db. For
         example if you use couchdbkit to access to cloudant or bigcouch:
-            
+
             Ex: q=12 or n=4
 
         See https://github.com/cloudant/bigcouch for more info.
 
         @return: Database instance if it's ok or dict message
         """
-        return Database(self._db_uri(dbname), create=True,
-                    server=self, **params)
+        return self.get_db(dbname, create=True, **params)
 
-    def get_or_create_db(self, dbname, **params):
-        """
+    get_or_create_db = create_db
+    get_or_create_db.__doc__ = """
         Try to return a Database object for dbname. If
         database doest't exist, it will be created.
 
         """
-        return Database(self._db_uri(dbname), create=True,
-                    server=self, **params)
 
     def delete_db(self, dbname):
         """
@@ -154,7 +164,7 @@ class Server(object):
         @param source: str, URI or dbname of the source
         @param target: str, URI or dbname of the target
         @param params: replication options
-        
+
         More info about replication here :
         http://wiki.apache.org/couchdb/Replication
 
@@ -166,7 +176,7 @@ class Server(object):
         payload.update(params)
         resp = self.res.post('/_replicate', payload=payload)
         return resp.json_body
-        
+
     def active_tasks(self):
         """ return active tasks """
         resp = self.res.get('/_active_tasks')
@@ -184,10 +194,11 @@ class Server(object):
         else:
             self._uuid_batch_count = self.uuid_batch_count
 
-        self.uuids = self.uuids or []
-        if not self._uuids:
-            self._uuids = self.uuids(count=self._uuid_batch_count)["uuids"]
-        return self._uuids.pop()
+        try:
+            return self._uuids.pop()
+        except IndexError:
+            self._uuids.extend(self.uuids(count=self._uuid_batch_count)["uuids"])
+            return self._uuids.pop()
 
     def __getitem__(self, dbname):
         return Database(self._db_uri(dbname), server=self)
@@ -217,7 +228,7 @@ class Server(object):
     def _db_uri(self, dbname):
         if dbname.startswith("/"):
             dbname = dbname[1:]
-            
+
         dbname = url_quote(dbname, safe=":")
         return "/".join([self.uri, dbname])
 
@@ -226,8 +237,7 @@ class Database(object):
     A Database object can act as a Dict object.
     """
 
-    def __init__(self, uri, create=False, server=None, pool_instance=None,
-            filters=None, **params):
+    def __init__(self, uri, create=False, server=None, **params):
         """Constructor for Database
 
         @param uri: str, Database uri
@@ -245,9 +255,7 @@ class Database(object):
                             server.__class__.__name__)
             self.server = server
         else:
-            self.server = server = Server(self.server_uri, 
-                                    pool_instance=pool_instance,
-                                    filters=filters)
+            self.server = server = Server(self.server_uri, **params)
 
         validate_dbname(self.dbname)
         if create:
@@ -256,7 +264,6 @@ class Database(object):
             except ResourceNotFound:
                 self.server.res.put('/%s/' % self.dbname, **params).json_body
 
-
         self.res = server.res(self.dbname)
 
     def __repr__(self):
@@ -279,7 +286,7 @@ class Database(object):
         path = "/_compact"
         if dname is not None:
             path = "%s/%s" % (path, resource.escape_docid(dname))
-        res = self.res.post(path, headers={"Content-Type": 
+        res = self.res.post(path, headers={"Content-Type":
             "application/json"})
         return res.json_body
 
@@ -346,7 +353,7 @@ class Database(object):
             wrapper = schema.wrap
 
         docid = resource.escape_docid(docid)
-        doc = self.res.get(docid, **params).json_body        
+        doc = self.res.get(docid, **params).json_body
         if wrapper is not None:
             if not callable(wrapper):
                 raise TypeError("wrapper isn't a callable")
@@ -356,6 +363,40 @@ class Database(object):
         return doc
     get = open_doc
 
+    def list(self, list_name, view_name, **params):
+        """ Execute a list function on the server and return the response.
+        If the response is json it will be deserialized, otherwise the string
+        will be returned.
+
+        Args:
+            @param list_name: should be 'designname/listname'
+            @param view_name: name of the view to run through the list document
+            @param params: params of the list
+        """
+        list_name = list_name.split('/')
+        dname = list_name.pop(0)
+        vname = '/'.join(list_name)
+        list_path = '_design/%s/_list/%s/%s' % (dname, vname, view_name)
+
+        return self.res.get(list_path, **params).json_body
+
+    def show(self, show_name, doc_id, **params):
+        """ Execute a show function on the server and return the response.
+        If the response is json it will be deserialized, otherwise the string
+        will be returned.
+
+        Args:
+            @param show_name: should be 'designname/showname'
+            @param doc_id: id of the document to pass into the show document
+            @param params: params of the show
+        """
+        show_name = show_name.split('/')
+        dname = show_name.pop(0)
+        vname = '/'.join(show_name)
+        show_path = '_design/%s/_show/%s/%s' % (dname, vname, doc_id)
+
+        return self.res.get(show_path, **params).json_body
+
     def all_docs(self, by_seq=False, **params):
         """Get all documents from a database
 
@@ -441,7 +482,7 @@ class Database(object):
             doc1.update({ '_id': res['id']})
         else:
             doc1.update({'_id': res['id'], '_rev': res['rev']})
-       
+
 
         if schema:
             doc._doc = doc1
@@ -449,7 +490,8 @@ class Database(object):
             doc.update(doc1)
         return res
 
-    def save_docs(self, docs, use_uuids=True, all_or_nothing=False):
+    def save_docs(self, docs, use_uuids=True, all_or_nothing=False,
+            **params):
         """ bulk save. Modify Multiple Documents With a Single Request
 
         @param docs: list of docs
@@ -461,7 +503,7 @@ class Database(object):
         .. seealso:: `HTTP Bulk Document API <http://wiki.apache.org/couchdb/HTTP_Bulk_Document_API>`
 
         """
-        
+
         docs1 = []
         docs_schema = []
         for doc in docs:
@@ -473,13 +515,10 @@ class Database(object):
             return '_id' in doc
 
         if use_uuids:
-            ids = []
             noids = []
             for k, g in groupby(docs1, is_id):
                 if not k:
                     noids = list(g)
-                else:
-                    ids = list(g)
 
             uuid_count = max(len(noids), self.server.uuid_batch_count)
             for doc in noids:
@@ -493,7 +532,7 @@ class Database(object):
 
         # update docs
         results = self.res.post('/_bulk_docs',
-                payload=payload).json_body
+                payload=payload, **params).json_body
 
         errors = []
         for i, res in enumerate(results):
@@ -515,19 +554,40 @@ class Database(object):
         return results
     bulk_save = save_docs
 
-    def delete_docs(self, docs, all_or_nothing=False):
+    def delete_docs(self, docs, all_or_nothing=False,
+            empty_on_delete=False, **params):
         """ bulk delete.
         It adds '_deleted' member to doc then uses bulk_save to save them.
 
+        @param empty_on_delete: default is False if you want to make
+        sure the doc is emptied and will not be stored as is in Apache
+        CouchDB.
+        @param all_or_nothing: In the case of a power failure, when the database
+        restarts either all the changes will have been saved or none of them.
+        However, it does not do conflict checking, so the documents will
+
+        .. seealso:: `HTTP Bulk Document API <http://wiki.apache.org/couchdb/HTTP_Bulk_Document_API>`
+
+
         """
-        for doc in docs:
-            doc['_deleted'] = True
+
+        if empty_on_delete:
+            for doc in docs:
+                new_doc = {"_id": doc["_id"],
+                        "_rev": doc["_rev"],
+                        "_deleted": True}
+                doc.clear()
+                doc.update(new_doc)
+        else:
+            for doc in docs:
+                doc['_deleted'] = True
+
         return self.bulk_save(docs, use_uuids=False,
-                all_or_nothing=all_or_nothing)
+                all_or_nothing=all_or_nothing, **params)
 
     bulk_delete = delete_docs
 
-    def delete_doc(self, doc):
+    def delete_doc(self, doc, **params):
         """ delete a document or a list of documents
         @param doc: str or dict,  document id or full doc.
         @return: dict like:
@@ -544,11 +604,11 @@ class Database(object):
                 raise KeyError('_id and _rev are required to delete a doc')
 
             docid = resource.escape_docid(doc1['_id'])
-            result = self.res.delete(docid, rev=doc1['_rev']).json_body
+            result = self.res.delete(docid, rev=doc1['_rev'], **params).json_body
         elif isinstance(doc1, basestring): # we get a docid
             rev = self.get_rev(doc1)
             docid = resource.escape_docid(doc1)
-            result = self.res.delete(docid, rev=rev).json_body
+            result = self.res.delete(docid, rev=rev, **params).json_body
 
         if schema:
             doc._doc.update({
@@ -562,12 +622,15 @@ class Database(object):
             })
         return result
 
-    def copy_doc(self, doc, dest=None):
+    def copy_doc(self, doc, dest=None, headers=None):
         """ copy an existing document to a new id. If dest is None, a new uuid will be requested
         @param doc: dict or string, document or document id
         @param dest: basestring or dict. if _rev is specified in dict it will override the doc
         """
 
+        if not headers:
+            headers = {}
+
         doc1, schema = _maybe_serialize(doc)
         if isinstance(doc1, basestring):
             docid = doc1
@@ -580,21 +643,19 @@ class Database(object):
             destination = self.server.next_uuid(count=1)
         elif isinstance(dest, basestring):
             if dest in self:
-                dest = self.get(dest)['_rev']
+                dest = self.get(dest)
                 destination = "%s?rev=%s" % (dest['_id'], dest['_rev'])
             else:
                 destination = dest
         elif isinstance(dest, dict):
             if '_id' in dest and '_rev' in dest and dest['_id'] in self:
-                rev = dest['_rev']
                 destination = "%s?rev=%s" % (dest['_id'], dest['_rev'])
             else:
                 raise KeyError("dest doesn't exist or this not a document ('_id' or '_rev' missig).")
 
         if destination:
-            result = self.res.copy('/%s' % docid, headers={ 
-                "Destination": str(destination)
-            }).json_body
+            headers.update({"Destination": str(destination)})
+            result = self.res.copy('/%s' % docid, headers=headers).json_body
             return result
 
         return { 'ok': False }
@@ -617,6 +678,36 @@ class Database(object):
         @param params: params of the view
 
         """
+        def get_multi_wrapper(classes, wrap_doc=True,
+                dynamic_properties=True):
+            def wrapper(row):
+                data = row.get('value')
+                docid = row.get('id')
+                doc = row.get('doc')
+                if doc is not None and wrap_doc:
+                    try:
+                        cls = classes[doc.get('doc_type')]
+                        cls._allow_dynamic_properties = dynamic_properties
+                        return cls.wrap(doc)
+                    except KeyError:
+                        return row
+                elif not data or data is None:
+                    return row
+                elif not isinstance(data, dict) or not docid:
+                    return row
+                else:
+                    try:
+                        cls = classes[data.get('doc_type')]
+                        data['_id'] = docid
+                        if 'rev' in data:
+                            data['_rev'] = data.pop('rev')
+                        cls._allow_dynamic_properties = dynamic_properties
+                        return cls.wrap(data)
+                    except KeyError:
+                        return row
+
+            return wrapper
+
         if view_name.startswith('/'):
             view_name = view_name[1:]
         if view_name == '_all_docs':
@@ -628,10 +719,24 @@ class Database(object):
             dname = view_name.pop(0)
             vname = '/'.join(view_name)
             view_path = '_design/%s/_view/%s' % (dname, vname)
+
         if schema is not None:
-            if not hasattr(schema, 'wrap'):
-                raise AttributeError(" no 'wrap' method found in obj %s)" % str(schemma))
-            wrapper = schema.wrap
+            wrap_doc = params.get('wrap_doc', True)
+            dynamic_properties = params.get('dynamic_properties', True)
+
+            if hasattr(schema, "wrap"):
+                if hasattr(schema, '_doc_type'):
+                    schema = {schema._doc_type: schema}
+                    wrapper = get_multi_wrapper(schema, wrap_doc=wrap_doc,
+                            dynamic_properties=dynamic_properties)
+                else:
+                    wrapper = schema.wrap
+            else:
+                if isinstance(schema, list):
+                    schema = dict([(s._doc_type, s) for s in schema])
+
+                wrapper = get_multi_wrapper(schema, wrap_doc=wrap_doc,
+                    dynamic_properties=dynamic_properties)
 
         return View(self, view_path, wrapper=wrapper)(**params)
 
@@ -648,15 +753,16 @@ class Database(object):
         with its default settings by default."""
         return View(self, "/%s/%s" % (handler, view_name), wrapper=wrapper)(**params)
 
-    def documents(self, wrapper=None, **params):
+    def documents(self, schema=None, wrapper=None, **params):
         """ return a ViewResults objects containing all documents.
         This is a shorthand to view function.
         """
-        return View(self, '_all_docs', wrapper=wrapper)(**params)
+        return View(self, '_all_docs', schema=schema,
+                wrapper=wrapper)(**params)
     iterdocuments = documents
 
     def put_attachment(self, doc, content, name=None, content_type=None,
-            content_length=None):
+            content_length=None, headers=None):
         """ Add attachement to a document. All attachments are streamed.
 
         @param doc: dict, document object
@@ -688,7 +794,8 @@ class Database(object):
             {u'ok': True}
         """
 
-        headers = {}
+        if not headers:
+            headers = {}
 
         if not content:
             content = ""
@@ -720,7 +827,7 @@ class Database(object):
             doc.update(new_doc)
         return res['ok']
 
-    def delete_attachment(self, doc, name):
+    def delete_attachment(self, doc, name, headers=None):
         """ delete attachement to the document
 
         @param doc: dict, document object in python
@@ -733,14 +840,16 @@ class Database(object):
         docid = resource.escape_docid(doc1['_id'])
         name = url_quote(name, safe="")
 
-        res = self.res(docid).delete(name, rev=doc1['_rev']).json_body
+        res = self.res(docid).delete(name, rev=doc1['_rev'],
+                headers=headers).json_body
         if res['ok']:
             new_doc = self.get(doc1['_id'], rev=res['rev'])
             doc.update(new_doc)
         return res['ok']
 
 
-    def fetch_attachment(self, id_or_doc, name, stream=False):
+    def fetch_attachment(self, id_or_doc, name, stream=False,
+            headers=None):
         """ get attachment in a document
 
         @param id_or_doc: str or dict, doc id or document dict
@@ -758,7 +867,7 @@ class Database(object):
         docid = resource.escape_docid(docid)
         name = url_quote(name, safe="")
 
-        resp = self.res(docid).get(name)
+        resp = self.res(docid).get(name, headers=headers)
         if stream:
             return resp.body_stream()
         return resp.body_string(charset="utf-8")
@@ -974,4 +1083,4 @@ class TempView(ViewInterface):
 
     def _exec(self, **params):
         return self._db.res.post('_temp_view', payload=self.design,
-                **params)
+            headers={"Content-Type": "application/json"}, **params)
